from http import HTTPStatus

from flask import Blueprint, redirect

from yak_server.database.models import TeamModel

from .utils.constants import GLOBAL_ENDPOINT, VERSION
from .utils.errors import InvalidTeamId, TeamNotFound
from .utils.flask_utils import is_iso_3166_1_alpha_2_code, is_uuid4, success_response

teams = Blueprint("team", __name__)


@teams.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/teams")
def teams_get():
    return success_response(
        HTTPStatus.OK,
        {"teams": [team.to_dict() for team in TeamModel.query.all()]},
    )


@teams.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/teams/<string:team_id>")
def teams_get_by_id(team_id):
    if is_uuid4(team_id):
        team = TeamModel.query.filter_by(id=team_id).first()
    elif is_iso_3166_1_alpha_2_code(team_id):
        team = TeamModel.query.filter_by(code=team_id).first()
    else:
        raise InvalidTeamId(team_id)

    if not team:
        raise TeamNotFound(team_id)

    return success_response(HTTPStatus.OK, {"team": team.to_dict()})


@teams.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/teams/<string:team_id>/flag")
def retrieve_team_flag(team_id):
    team = TeamModel.query.filter_by(id=team_id).first()

    if not team:
        raise TeamNotFound(team_id)

    return redirect(team.flag_url)



import json
import logging
import subprocess
import sys
from datetime import datetime
from getpass import getpass

if sys.version_info >= (3, 9):
    from importlib import resources
else:
    import importlib_resources as resources

from pathlib import Path

from yak_server import db
from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    GroupPositionModel,
    MatchModel,
    PhaseModel,
    ScoreBetModel,
    TeamModel,
    UserModel,
)

logger = logging.getLogger(__name__)


class ConfirmPasswordDoesNotMatch(Exception):
    def __init__(self) -> None:
        super().__init__("Password and Confirm Password fields does not match.")


class SignupError(Exception):
    def __init__(self, description) -> None:
        super().__init__(f"Error during signup. {description}")


class RecordDeletionInProduction(Exception):
    def __init__(self) -> None:
        super().__init__("Trying to delete records in production using script. DO NOT DO IT.")


class TableDropInProduction(Exception):
    def __init__(self) -> None:
        super().__init__("Trying to drop database tables in production using script. DO NOT DO IT.")


class BackupError(Exception):
    def __init__(self, description) -> None:
        super().__init__(f"Error during backup. {description}")


def create_database(app):
    db.create_all()


def create_admin(app):
    password = getpass(prompt="Admin user password: ")
    confirm_password = getpass(prompt="Confirm admin password: ")

    if password != confirm_password:
        raise ConfirmPasswordDoesNotMatch

    client = app.test_client()

    response_signup = client.post(
        "/api/v1/users/signup",
        json={
            "name": "admin",
            "first_name": "admin",
            "last_name": "admin",
            "password": password,
        },
    )

    if not response_signup.json["ok"]:
        raise SignupError(response_signup.json["description"])


def initialize_database(app):
    DATA_FOLDER = app.config["DATA_FOLDER"]

    with Path(f"{DATA_FOLDER}/phases.json").open() as file:
        phases = json.loads(file.read())

        db.session.add_all(PhaseModel(**phase) for phase in phases)
        db.session.flush()

    with Path(f"{DATA_FOLDER}/groups.json").open() as file:
        groups = json.loads(file.read())

        for group in groups:
            phase = PhaseModel.query.filter_by(code=group["phase_code"]).first()
            group.pop("phase_code")
            group["phase_id"] = phase.id

        db.session.add_all(GroupModel(**group) for group in groups)
        db.session.flush()

    with Path(f"{DATA_FOLDER}/teams.json").open() as file:
        teams = json.loads(file.read())

        db.session.add_all(TeamModel(**team) for team in teams)
        db.session.flush()

    with Path(f"{DATA_FOLDER}/matches.json").open() as file:
        matches = json.loads(file.read())

        for match in matches:
            team1 = TeamModel.query.filter_by(code=match["team1_code"]).first()
            match.pop("team1_code")
            match["team1_id"] = team1.id

            team2 = TeamModel.query.filter_by(code=match["team2_code"]).first()
            match.pop("team2_code")
            match["team2_id"] = team2.id

            group = GroupModel.query.filter_by(code=match["group_code"]).first()
            match.pop("group_code")
            match["group_id"] = group.id

        db.session.add_all(MatchModel(**match) for match in matches)
        db.session.flush()

    db.session.commit()


def backup_database(app):
    with resources.as_file(resources.files("yak_server") / "cli/backup_files") as path:
        backup_location = path

    if not Path(backup_location).exists():
        Path(backup_location).mkdir()

    backup_date, backup_time = datetime.now().strftime("%d-%m-%Y %H:%M:%S").split()

    file_name = f"{backup_location}/yak_toto_backup_{backup_date}T{backup_time}.sql"

    result = subprocess.run(
        (
            f"mysqldump {app.config['MYSQL_DB']} "
            f"-u {app.config['MYSQL_USER_NAME']} "
            f"-P {app.config['MYSQL_PORT']} "
            "--protocol=tcp "
            f"--password={app.config['MYSQL_PASSWORD']}"
        ),
        shell=True,
        capture_output=True,
        encoding="utf-8",
    )

    if result.returncode:
        error_message = (
            f"Something went wrong when backup on {backup_date} at {backup_time}: "
            f"{result.stderr.replace(app.config['MYSQL_PASSWORD'], '********')}"
        )

        logger.error(error_message)

        raise BackupError(error_message)

    with Path(file_name).open(mode="w") as file:
        file.write(result.stdout)
        logger.info(f"Backup done on {backup_date} at {backup_time}")


def delete_database(app):
    if not app.config.get("DEBUG"):
        raise RecordDeletionInProduction

    GroupPositionModel.query.delete()
    ScoreBetModel.query.delete()
    BinaryBetModel.query.delete()
    UserModel.query.delete()
    MatchModel.query.delete()
    GroupModel.query.delete()
    PhaseModel.query.delete()
    TeamModel.query.delete()
    db.session.commit()


def drop_database(app):
    if not app.config.get("DEBUG"):
        raise TableDropInProduction

    db.drop_all()



import logging
import sys

if sys.version_info >= (3, 9):
    from importlib import resources
else:
    import importlib_resources as resources

from flask import Flask
from flask.cli import load_dotenv
from flask_cors import CORS
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy
from strawberry.flask.views import GraphQLView

db = SQLAlchemy()

logger = logging.getLogger(__name__)

load_dotenv()


def create_app():
    app = Flask(__name__)

    with resources.as_file(resources.files("yak_server") / "database/migrations") as path:
        Migrate(app, db, directory=path)

    # Configuration setup
    from .config_file import get_config

    app.config.from_mapping(get_config())
    app.json.sort_keys = False

    # Database setup
    db.init_app(app)

    # --------------------------------------------- #
    # Version 1 setup (REST api)
    # --------------------------------------------- #

    # Registrer error handler
    from .v1.utils.errors import set_error_handler

    set_error_handler(app)

    # Registrer blueprint
    from .v1.auth import auth as auth_blueprint
    from .v1.bets import bets as bets_blueprint
    from .v1.binary_bets import binary_bets as binary_bets_blueprint
    from .v1.groups import groups as groups_blueprint
    from .v1.phase import phase as phase_blueprint
    from .v1.results import results as results_blueprint
    from .v1.rules import rules as rules_blueprint
    from .v1.score_bets import score_bets as score_bets_blueprint
    from .v1.teams import teams as teams_blueprint

    app.register_blueprint(auth_blueprint)
    app.register_blueprint(bets_blueprint)
    app.register_blueprint(binary_bets_blueprint)
    app.register_blueprint(groups_blueprint)
    app.register_blueprint(phase_blueprint)
    app.register_blueprint(results_blueprint)
    app.register_blueprint(rules_blueprint)
    app.register_blueprint(score_bets_blueprint)
    app.register_blueprint(teams_blueprint)

    # --------------------------------------------- #
    # Version 2 setup (GraphQL api)
    # --------------------------------------------- #
    from .v2 import schema
    from .v2.utils.constants import GLOBAL_ENDPOINT, VERSION

    # Registrer endpoint
    app.add_url_rule(
        f"/{GLOBAL_ENDPOINT}/{VERSION}",
        view_func=GraphQLView.as_view("graphql_view", schema=schema, graphiql=app.config["DEBUG"]),
    )

    CORS(app, resources={r"/api/*": {"origins": "*"}})

    # --------------------------------------------- #
    # Declare logger configuration for yak server
    # --------------------------------------------- #
    logging.basicConfig(
        filename="yak.log",
        level=logging.DEBUG if app.config.get("DEBUG") else logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    logger.info("Start yak flask server")

    return app



from functools import wraps

from flask import current_app, request
from jwt import ExpiredSignatureError

from yak_server.database.models import UserModel
from yak_server.helpers.authentification import decode_bearer_token

from .errors import ExpiredToken, InvalidToken, UnauthorizedAccessToAdminAPI, UserNotFound

NUMBER_ELEMENTS_IN_AUTHORIZATION = 2


def user_from_token(auth_headers):
    if len(auth_headers) != NUMBER_ELEMENTS_IN_AUTHORIZATION or auth_headers[0] != "Bearer":
        raise InvalidToken

    token = auth_headers[1]
    try:
        data = decode_bearer_token(token, current_app.config["SECRET_KEY"])
    except ExpiredSignatureError as exc:
        raise ExpiredToken from exc
    except Exception as exc:
        raise InvalidToken from exc

    user = UserModel.query.filter_by(id=data["sub"]).first()
    if not user:
        raise UserNotFound

    return user


def is_authentificated(f):
    @wraps(f)
    def _verify(*args, **kwargs):
        auth_headers = request.headers.get("Authorization", "").split()

        user = user_from_token(auth_headers)

        return f(user, *args, **kwargs)

    return _verify


def is_admin_authentificated(f):
    @wraps(f)
    def _verify(*args, **kwargs):
        auth_headers = request.headers.get("Authorization", "").split()

        user = user_from_token(auth_headers)

        if user.name != "admin":
            raise UnauthorizedAccessToAdminAPI

        return f(user, *args, **kwargs)

    return _verify



import logging
from http import HTTPStatus

from flask import Blueprint, request
from sqlalchemy.exc import IntegrityError

from yak_server import db
from yak_server.database.models import (
    BinaryBetModel,
    MatchModel,
    is_locked,
)
from yak_server.helpers.logging import modify_binary_bet_successfully

from .utils.auth_utils import is_authentificated
from .utils.constants import GLOBAL_ENDPOINT, VERSION
from .utils.errors import (
    BetNotFound,
    GroupNotFound,
    LockedBinaryBet,
    TeamNotFound,
)
from .utils.flask_utils import success_response
from .utils.schemas import SCHEMA_PATCH_BINARY_BET, SCHEMA_POST_BINARY_BET
from .utils.validation import validate_body

binary_bets = Blueprint("binary_bets", __name__)

logger = logging.getLogger(__name__)


@binary_bets.post(f"/{GLOBAL_ENDPOINT}/{VERSION}/binary_bets")
@validate_body(schema=SCHEMA_POST_BINARY_BET)
@is_authentificated
def create_binary_bet(user):
    if is_locked(user.name):
        raise LockedBinaryBet

    body = request.get_json()

    match = MatchModel(
        team1_id=body["team1"]["id"],
        team2_id=body["team2"]["id"],
        index=body["index"],
        group_id=body["group"]["id"],
    )

    db.session.add(match)

    try:
        db.session.flush()
    except IntegrityError as integrity_error:
        if "FOREIGN KEY (`team1_id`)" in str(integrity_error):
            raise TeamNotFound(team_id=body["team1"]["id"]) from integrity_error
        elif "FOREIGN KEY (`team2_id`)" in str(integrity_error):
            raise TeamNotFound(team_id=body["team2"]["id"]) from integrity_error
        else:
            raise GroupNotFound(group_id=body["group"]["id"]) from integrity_error

    binary_bet = BinaryBetModel(
        match_id=match.id,
        user_id=user.id,
        is_one_won=body.get("is_one_won"),
    )

    db.session.add(binary_bet)
    db.session.commit()

    return success_response(
        HTTPStatus.CREATED,
        {
            "phase": binary_bet.match.group.phase.to_dict(),
            "group": binary_bet.match.group.to_dict_without_phase(),
            "binary_bet": binary_bet.to_dict_without_group(),
        },
    )


@binary_bets.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/binary_bets/<string:bet_id>")
@is_authentificated
def retrieve_binary_bet(user, bet_id):
    binary_bet = BinaryBetModel.query.filter_by(user_id=user.id, id=bet_id).first()

    if not binary_bet:
        raise BetNotFound(bet_id)

    return success_response(
        HTTPStatus.OK,
        {
            "phase": binary_bet.match.group.phase.to_dict(),
            "group": binary_bet.match.group.to_dict_without_phase(),
            "binary_bet": binary_bet.to_dict_without_group(),
        },
    )


@binary_bets.patch(f"/{GLOBAL_ENDPOINT}/{VERSION}/binary_bets/<string:bet_id>")
@validate_body(schema=SCHEMA_PATCH_BINARY_BET)
@is_authentificated
def modify_binary_bet(user, bet_id):
    if is_locked(user.name):
        raise LockedBinaryBet

    binary_bet = BinaryBetModel.query.filter_by(user_id=user.id, id=bet_id).first()

    if not binary_bet:
        raise BetNotFound(bet_id)

    body = request.get_json()

    logger.info(modify_binary_bet_successfully(user.name, binary_bet, body["is_one_won"]))

    binary_bet.is_one_won = body["is_one_won"]
    db.session.commit()

    return success_response(
        HTTPStatus.OK,
        {
            "phase": binary_bet.match.group.phase.to_dict(),
            "group": binary_bet.match.group.to_dict_without_phase(),
            "binary_bet": binary_bet.to_dict_without_group(),
        },
    )


@binary_bets.delete(f"/{GLOBAL_ENDPOINT}/{VERSION}/binary_bets/<string:bet_id>")
@is_authentificated
def delete_binary_bet(user, bet_id):
    if is_locked(user.name):
        raise LockedBinaryBet

    binary_bet = BinaryBetModel.query.filter_by(id=bet_id, user_id=user.id).first()

    if not binary_bet:
        raise BetNotFound(bet_id)

    response_body = {
        "phase": binary_bet.match.group.phase.to_dict(),
        "group": binary_bet.match.group.to_dict_without_phase(),
        "binary_bet": binary_bet.to_dict_without_group(),
    }

    db.session.delete(binary_bet)
    db.session.commit()

    return success_response(HTTPStatus.OK, response_body)



from itertools import chain
from operator import attrgetter
from uuid import uuid4

from sqlalchemy import and_

from yak_server import db
from yak_server.database.models import BinaryBetModel, GroupModel, MatchModel, PhaseModel
from yak_server.v1.bets import get_group_rank_with_code


def compute_finale_phase_from_group_rank(user, rule_config):
    groups_result = {
        group.code: get_group_rank_with_code(user, group.code)["group_rank"]
        for group in GroupModel.query.join(GroupModel.phase).filter(
            PhaseModel.code == rule_config["first_phase"],
        )
    }

    first_phase_phase_group = GroupModel.query.filter_by(
        code=rule_config["first_group"],
    ).first()

    existing_binary_bets = BinaryBetModel.query.join(BinaryBetModel.match).filter(
        and_(
            BinaryBetModel.user_id == user.id,
            MatchModel.group_id == first_phase_phase_group.id,
        ),
    )

    new_binary_bets = []
    new_matches = []

    for index, match_config in enumerate(rule_config["versus"], 1):
        if all(
            team["played"] == len(groups_result[match_config["team1"]["group"]]) - 1
            for team in chain(
                groups_result[match_config["team1"]["group"]],
                groups_result[match_config["team2"]["group"]],
            )
        ):
            team1 = groups_result[match_config["team1"]["group"]][
                match_config["team1"]["rank"] - 1
            ]["team"]
            team2 = groups_result[match_config["team2"]["group"]][
                match_config["team2"]["rank"] - 1
            ]["team"]

            match = MatchModel.query.filter_by(
                group_id=first_phase_phase_group.id,
                team1_id=team1["id"],
                team2_id=team2["id"],
                index=index,
            ).first()

            if not match:
                match = MatchModel(
                    id=str(uuid4()),
                    group_id=first_phase_phase_group.id,
                    team1_id=team1["id"],
                    team2_id=team2["id"],
                    index=index,
                )

            new_matches.append(match)

            bet = BinaryBetModel.query.filter_by(
                user_id=user.id,
                match_id=match.id,
            ).first()

            if not bet:
                bet = BinaryBetModel(
                    user_id=user.id,
                    match_id=match.id,
                )

            new_binary_bets.append(bet)

    # Compare existing matches/bets and new matches/bets
    db.session.add_all(new_matches)
    db.session.flush()

    db.session.add_all(new_binary_bets)
    db.session.flush()

    is_bet_modified = False

    for bet in existing_binary_bets:
        if bet.id not in map(attrgetter("id"), new_binary_bets):
            is_bet_modified = True
            db.session.delete(bet)

    db.session.flush()

    if is_bet_modified:
        for bet in BinaryBetModel.query.filter_by(user_id=user.id):
            if bet.match.group.code in GroupModel.query.join(GroupModel.phase).filter(
                and_(
                    PhaseModel.code == rule_config["second_phase"],
                    GroupModel.id == first_phase_phase_group.id,
                ),
            ):
                db.session.delete(bet)
        db.session.flush()

    db.session.commit()


RULE_MAPPING = {
    "492345de-8d4a-45b6-8b94-d219f2b0c3e9": compute_finale_phase_from_group_rank,
}



from http import HTTPStatus
from itertools import chain

from flask import Blueprint, current_app
from sqlalchemy import and_

from yak_server import db
from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    MatchModel,
    PhaseModel,
    ScoreBetModel,
    UserModel,
)

from .bets import get_group_rank_with_code
from .utils.auth_utils import is_admin_authentificated, is_authentificated
from .utils.constants import GLOBAL_ENDPOINT, VERSION
from .utils.errors import NoResultsForAdminUser
from .utils.flask_utils import success_response

results = Blueprint("results", __name__)


@results.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/score_board")
@is_authentificated
def score_board(current_user):
    return success_response(
        HTTPStatus.OK,
        [
            user.to_result_dict()
            for user in UserModel.query.order_by(UserModel.points.desc()).filter(
                UserModel.name != "admin",
            )
        ],
    )


@results.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/results")
@is_authentificated
def results_get(current_user):
    if current_user.name == "admin":
        raise NoResultsForAdminUser

    results = UserModel.query.order_by(UserModel.points.desc()).filter(
        UserModel.name != "admin",
    )

    rank = [
        index
        for index, user_result in enumerate(results, start=1)
        if user_result.id == current_user.id
    ][0]

    return success_response(
        HTTPStatus.OK,
        {**UserModel.query.filter_by(id=current_user.id).first().to_result_dict(), "rank": rank},
    )


@results.post(f"/{GLOBAL_ENDPOINT}/{VERSION}/compute_points")
@is_admin_authentificated
def compute_points_post(_):
    compute_points(
        current_app.config["BASE_CORRECT_RESULT"],
        current_app.config["MULTIPLYING_FACTOR_CORRECT_RESULT"],
        current_app.config["BASE_CORRECT_SCORE"],
        current_app.config["MULTIPLYING_FACTOR_CORRECT_SCORE"],
        current_app.config["TEAM_QUALIFIED"],
        current_app.config["FIRST_TEAM_QUALIFIED"],
    )

    return success_response(
        HTTPStatus.OK,
        [
            user.to_result_dict()
            for user in UserModel.query.order_by(UserModel.points.desc()).filter(
                UserModel.name != "admin",
            )
        ],
    )


def compute_points(
    base_correct_result,
    multiplying_factor_correct_result,
    base_correct_score,
    multiplying_factor_correct_score,
    team_qualified,
    first_team_qualified,
):
    admin = UserModel.query.filter_by(name="admin").first()

    results = []

    for real_score in admin.score_bets:
        number_correct_result = 0
        number_correct_score = 0
        user_ids_found_correct_result = []
        user_ids_found_correct_score = []

        for user_score in ScoreBetModel.query.filter(
            and_(
                ScoreBetModel.match_id == real_score.match_id,
                ScoreBetModel.user_id != admin.id,
            ),
        ):
            if user_score.is_same_results(real_score):
                number_correct_result += 1
                user_ids_found_correct_result.append(user_score.user_id)

                if user_score.is_same_scores(real_score):
                    number_correct_score += 1
                    user_ids_found_correct_score.append(user_score.user_id)

        results.append(
            {
                "match_id": real_score.match_id,
                "number_correct_result": number_correct_result,
                "user_ids_found_correct_result": user_ids_found_correct_result,
                "number_correct_score": number_correct_score,
                "user_ids_found_correct_score": user_ids_found_correct_score,
            },
        )

    result_groups = {}

    users = UserModel.query.filter(UserModel.name != "admin")

    for group in GroupModel.query.join(GroupModel.phase).filter(
        PhaseModel.code == "GROUP",
    ):
        group_result_admin = get_group_rank_with_code(admin, group.code)["group_rank"]

        if all_results_filled_in_group(group_result_admin):
            admin_first_team_id = group_result_admin[0]["team"]["id"]
            admin_second_team_id = group_result_admin[1]["team"]["id"]

            for user in users:
                if user.id not in result_groups:
                    result_groups[user.id] = {
                        "number_qualified_teams_guess": 0,
                        "number_first_qualified_guess": 0,
                    }

                group_result_user = get_group_rank_with_code(user, group.code)["group_rank"]

                if all_results_filled_in_group(group_result_user):
                    user_first_team_id = group_result_user[0]["team"]["id"]
                    user_second_team_id = group_result_user[1]["team"]["id"]

                    result_groups[user.id]["number_qualified_teams_guess"] += len(
                        {user_first_team_id, user_second_team_id}
                        & {admin_first_team_id, admin_second_team_id},
                    )

                    if user_first_team_id == admin_first_team_id:
                        result_groups[user.id]["number_first_qualified_guess"] += 1

    quarter_finals_team = team_from_group_code(admin, "4")
    semi_finals_team = team_from_group_code(admin, "2")
    final_team = team_from_group_code(admin, "1")
    winner = winner_from_user(admin)

    numbers_of_players = users.count()

    for user in users:
        user.number_score_guess = 0
        user.number_match_guess = 0
        user.points = 0

        for result in results:
            if user.id in result["user_ids_found_correct_result"]:
                user.number_match_guess += 1
                user.points += base_correct_result + multiplying_factor_correct_result * (
                    numbers_of_players - result["number_correct_result"]
                ) / (numbers_of_players - 1)

            if user.id in result["user_ids_found_correct_score"]:
                user.number_score_guess += 1
                user.points += base_correct_score + multiplying_factor_correct_score * (
                    numbers_of_players - result["number_correct_score"]
                ) / (numbers_of_players - 1)

        user.number_qualified_teams_guess = result_groups.get(user.id, {}).get(
            "number_qualified_teams_guess",
            0,
        )
        user.number_first_qualified_guess = result_groups.get(user.id, {}).get(
            "number_first_qualified_guess",
            0,
        )

        user.points += user.number_qualified_teams_guess * team_qualified
        user.points += user.number_first_qualified_guess * first_team_qualified

        user.number_quarter_final_guess = len(
            team_from_group_code(user, "4").intersection(quarter_finals_team),
        )
        user.number_semi_final_guess = len(
            team_from_group_code(user, "2").intersection(semi_finals_team),
        )
        user.number_final_guess = len(
            team_from_group_code(user, "1").intersection(final_team),
        )
        user.number_winner_guess = len(winner_from_user(user).intersection(winner))

        user.points += 30 * user.number_quarter_final_guess
        user.points += 60 * user.number_semi_final_guess
        user.points += 120 * user.number_final_guess
        user.points += 200 * user.number_winner_guess

    db.session.commit()


def all_results_filled_in_group(group_result):
    return all(team["played"] == len(group_result) - 1 for team in group_result)


def team_from_group_code(user, group_code):
    return set(
        chain(
            *(
                (bet.match.team1.id, bet.match.team2.id)
                for bet in user.binary_bets.filter(GroupModel.code == group_code)
                .join(BinaryBetModel.match)
                .join(MatchModel.group)
            ),
        ),
    )


def winner_from_user(user):
    finale_bet = list(
        user.binary_bets.filter(GroupModel.code == "1")
        .join(BinaryBetModel.match)
        .join(MatchModel.group),
    )

    if not finale_bet:
        return set()

    finale_bet = finale_bet[0]

    if finale_bet.is_one_won is None:
        return set()

    return {
        finale_bet.match.team1.id if finale_bet.is_one_won else finale_bet.match.team2.id,
    }



import logging
import traceback
from http import HTTPStatus

from flask import Response, jsonify
from werkzeug.exceptions import HTTPException

from yak_server.helpers.errors import (
    EXPIRED_TOKEN_MESSAGE,
    INVALID_CREDENTIALS_MESSAGE,
    INVALID_TOKEN_MESSAGE,
    LOCKED_BINARY_BET_MESSAGE,
    LOCKED_SCORE_BET_MESSAGE,
    UNAUTHORIZED_ACCESS_TO_ADMIN_API_MESSAGE,
    name_already_exists_message,
)

logger = logging.getLogger(__name__)


class InvalidCredentials(HTTPException):
    code = HTTPStatus.UNAUTHORIZED
    description = INVALID_CREDENTIALS_MESSAGE


class NameAlreadyExists(HTTPException):
    def __init__(self, name) -> None:
        super().__init__(name_already_exists_message(name))
        self.code = HTTPStatus.UNAUTHORIZED


class BetNotFound(HTTPException):
    def __init__(self, bet_id) -> None:
        super().__init__(f"Bet not found: {bet_id}")
        self.code = HTTPStatus.NOT_FOUND


class UserNotFound(HTTPException):
    code = HTTPStatus.NOT_FOUND
    description = "User not found"


class UnauthorizedAccessToAdminAPI(HTTPException):
    code = HTTPStatus.UNAUTHORIZED
    description = UNAUTHORIZED_ACCESS_TO_ADMIN_API_MESSAGE


class InvalidTeamId(HTTPException):
    def __init__(self, team_id):
        super().__init__(
            f"Invalid team id: {team_id}. Retry with a uuid or ISO 3166-1 alpha-2 code",
        )
        self.code = HTTPStatus.BAD_REQUEST


class TeamNotFound(HTTPException):
    def __init__(self, team_id):
        super().__init__(f"Team not found: {team_id}")
        self.code = HTTPStatus.NOT_FOUND


class LockedScoreBet(HTTPException):
    code = HTTPStatus.UNAUTHORIZED
    description = LOCKED_SCORE_BET_MESSAGE


class LockedBinaryBet(HTTPException):
    code = HTTPStatus.UNAUTHORIZED
    description = LOCKED_BINARY_BET_MESSAGE


class NoResultsForAdminUser(HTTPException):
    code = HTTPStatus.UNAUTHORIZED
    description = "No results for admin user"


class GroupNotFound(HTTPException):
    def __init__(self, group_id) -> None:
        super().__init__(f"Group not found: {group_id}")
        self.code = HTTPStatus.NOT_FOUND


class PhaseNotFound(HTTPException):
    def __init__(self, phase_id) -> None:
        super().__init__(f"Phase not found: {phase_id}")
        self.code = HTTPStatus.NOT_FOUND


class InvalidToken(HTTPException):
    code = HTTPStatus.UNAUTHORIZED
    description = INVALID_TOKEN_MESSAGE


class ExpiredToken(HTTPException):
    code = HTTPStatus.UNAUTHORIZED
    description = EXPIRED_TOKEN_MESSAGE


class RuleNotFound(HTTPException):
    def __init__(self, rule_id) -> None:
        super().__init__(f"Rule not found: {rule_id}")
        self.code = HTTPStatus.NOT_FOUND


class RequestValidationError(HTTPException):
    def __init__(self, schema, path, description):
        super().__init__(description)
        self.schema = schema
        self.path = path
        self.code = HTTPStatus.UNPROCESSABLE_ENTITY


def set_error_handler(app):
    @app.errorhandler(HTTPException)
    def handle_http_exception(e: HTTPException) -> Response:
        # Return JSON instead of HTML for HTTP errors.
        logger.info(f"Server catches an expected exception: {type(e).__name__} {e.description}")

        return jsonify(ok=False, error_code=e.code, description=e.description), e.code

    @app.errorhandler(Exception)
    def handle_exception(e: Exception) -> Response:
        # Return JSON instead of HTML for generic errors.
        logger.error(traceback.format_exc())
        logger.error(f"An unexcepted expection occurs: {type(e).__name__} {e}")

        return (
            jsonify(
                {
                    "ok": False,
                    "error_code": HTTPStatus.INTERNAL_SERVER_ERROR,
                    "description": f"{type(e).__name__}: {str(e)}"
                    if app.config.get("DEBUG")
                    else "Unexcepted error",
                },
            ),
            HTTPStatus.INTERNAL_SERVER_ERROR,
        )

    @app.errorhandler(RequestValidationError)
    def handle_request_validation_error(e: RequestValidationError) -> Response:
        return (
            jsonify(
                {
                    "ok": False,
                    "error_code": e.code,
                    "description": e.description,
                    "schema": e.schema,
                    "path": list(e.path),
                },
            ),
            e.code,
        )



from functools import wraps

from flask import request
from jsonschema import Draft202012Validator, ValidationError, validate

from .errors import RequestValidationError


def validate_body(schema):
    def decorator(f):
        @wraps(f)
        def _verify(*args, **kwargs):
            try:
                validate(
                    instance=request.get_json(),
                    schema=schema,
                    format_checker=Draft202012Validator.FORMAT_CHECKER,
                )
            except ValidationError as validation_error:
                raise RequestValidationError(
                    schema=validation_error.schema,
                    description=validation_error.message,
                    path=list(validation_error.absolute_path),
                ) from validation_error

            return f(*args, **kwargs)

        return _verify

    return decorator



import logging
from datetime import timedelta
from typing import Optional
from uuid import UUID

import strawberry
from flask import current_app
from sqlalchemy import sql, update
from strawberry.types import Info

from yak_server import db
from yak_server.database.models import (
    BinaryBetModel,
    GroupPositionModel,
    MatchModel,
    ScoreBetModel,
    UserModel,
    is_locked,
)
from yak_server.helpers.authentification import encode_bearer_token
from yak_server.helpers.group_position import create_group_position
from yak_server.helpers.logging import (
    logged_in_successfully,
    modify_binary_bet_successfully,
    modify_score_bet_successfully,
    signed_up_successfully,
)

from .bearer_authenfication import (
    is_admin_authentificated,
    is_authentificated,
)
from .result import (
    BinaryBetNotFoundForUpdate,
    InvalidCredentials,
    LockedBinaryBetError,
    LockedScoreBetError,
    LoginResult,
    ModifyBinaryBetResult,
    ModifyScoreBetResult,
    ModifyUserResult,
    NewScoreNegative,
    ScoreBetNotFoundForUpdate,
    SignupResult,
    UserNameAlreadyExists,
    UserNotFound,
    UserWithoutSensitiveInfo,
)
from .schema import (
    BinaryBet,
    ScoreBet,
    UserWithToken,
)

logger = logging.getLogger(__name__)


@strawberry.type
class Mutation:
    @strawberry.mutation
    def signup_result(
        self,
        user_name: str,
        password: str,
        first_name: str,
        last_name: str,
    ) -> SignupResult:
        # Check existing user in db
        existing_user = UserModel.query.filter_by(name=user_name).first()
        if existing_user:
            return UserNameAlreadyExists(user_name=user_name)

        # Initialize user and integrate in db
        user = UserModel(
            name=user_name,
            first_name=first_name,
            last_name=last_name,
            password=password,
        )
        db.session.add(user)
        db.session.flush()

        # Initialize bets and integrate in db
        db.session.add_all(
            match.bet_type_from_match.value(user_id=user.id, match_id=match.id)
            for match in MatchModel.query.filter(MatchModel.bet_type_from_match is not sql.null())
        )
        db.session.flush()

        # Create group position records
        db.session.add_all(create_group_position(ScoreBetModel.query.filter_by(user_id=user.id)))
        db.session.commit()

        token = encode_bearer_token(
            sub=user.id,
            expiration_time=timedelta(seconds=current_app.config["JWT_EXPIRATION_TIME"]),
            secret_key=current_app.config["SECRET_KEY"],
        )

        logger.info(signed_up_successfully(user.name))

        return UserWithToken.from_instance(instance=user, token=token)

    @strawberry.mutation
    def login_result(self, user_name: str, password: str) -> LoginResult:
        user = UserModel.authenticate(name=user_name, password=password)

        if not user:
            return InvalidCredentials()

        token = encode_bearer_token(
            sub=user.id,
            expiration_time=timedelta(seconds=current_app.config["JWT_EXPIRATION_TIME"]),
            secret_key=current_app.config["SECRET_KEY"],
        )

        logger.info(logged_in_successfully(user.name))

        return UserWithToken.from_instance(instance=user, token=token)

    @strawberry.mutation
    @is_authentificated
    def modify_binary_bet_result(
        self,
        id: UUID,
        is_one_won: Optional[bool],
        info: Info,
    ) -> ModifyBinaryBetResult:
        bet = BinaryBetModel.query.filter_by(user_id=info.user.instance.id, id=str(id)).first()

        if is_locked(info.user.pseudo):
            return LockedBinaryBetError()

        if not bet:
            return BinaryBetNotFoundForUpdate()

        logger.info(modify_binary_bet_successfully(info.user.pseudo, bet, is_one_won))

        bet.is_one_won = is_one_won
        db.session.commit()

        return BinaryBet.from_instance(instance=bet)

    @strawberry.mutation
    @is_authentificated
    def modify_score_bet_result(
        self,
        id: UUID,
        score1: Optional[int],
        score2: Optional[int],
        info: Info,
    ) -> ModifyScoreBetResult:
        bet = ScoreBetModel.query.filter_by(user_id=info.user.instance.id, id=str(id)).first()

        if is_locked(info.user.pseudo):
            return LockedScoreBetError()

        if not bet:
            return ScoreBetNotFoundForUpdate()

        if score1 is not None and score1 < 0:
            return NewScoreNegative(variable_name="$score1", score=score1)

        if score2 is not None and score2 < 0:
            return NewScoreNegative(variable_name="$score2", score=score2)

        if score1 == bet.score1 and score2 == bet.score2:
            return ScoreBet.from_instance(instance=bet)

        db.session.execute(
            update(GroupPositionModel)
            .values(need_recomputation=True)
            .where(
                GroupPositionModel.team_id == bet.match.team1_id,
                GroupPositionModel.user_id == info.user.id,
            ),
        )
        db.session.execute(
            update(GroupPositionModel)
            .values(need_recomputation=True)
            .where(
                GroupPositionModel.team_id == bet.match.team2_id,
                GroupPositionModel.user_id == info.user.id,
            ),
        )
        logger.info(modify_score_bet_successfully(info.user.pseudo, bet, score1, score2))

        bet.score1 = score1
        bet.score2 = score2
        db.session.commit()

        return ScoreBet.from_instance(instance=bet)

    @strawberry.mutation
    @is_authentificated
    @is_admin_authentificated
    def modify_user_result(self, id: UUID, password: str, info: Info) -> ModifyUserResult:
        user = UserModel.query.filter_by(id=str(id)).first()

        if not user:
            return UserNotFound(id=id)

        user.change_password(password)
        db.session.commit()

        return UserWithoutSensitiveInfo.from_instance(instance=user)



def modify_score_bet_successfully(user_name, original_bet, new_score1, new_score2) -> str:
    return (
        f"{user_name} modify "
        f"{original_bet.match.team1.description}-{original_bet.match.team2.description} "
        f"in {original_bet.match.group.description} "
        f"from {original_bet.score1}-{original_bet.score2} "
        f"to {new_score1}-{new_score2}"
    )


def modify_binary_bet_successfully(user_name, original_bet, new_is_one_won) -> str:
    return (
        f"{user_name} modify "
        f"{original_bet.match.team1.description}-{original_bet.match.team2.description} "
        f"in {original_bet.match.group.description} "
        f"from {original_bet.is_one_won}-"
        f"{not original_bet.is_one_won if isinstance(original_bet.is_one_won, bool) else None} "
        f"to {new_is_one_won}-{not new_is_one_won if isinstance(new_is_one_won, bool) else None}"
    )


def signed_up_successfully(user_name):
    return f"{user_name} signed up successfully"


def logged_in_successfully(user_name):
    return f"{user_name} logged in successfully"


def modify_password_successfully(user_name):
    return f"admin user modify {user_name} password"



import logging
from datetime import timedelta
from http import HTTPStatus

from flask import Blueprint, current_app, request
from sqlalchemy import sql

from yak_server import db
from yak_server.database.models import MatchModel, ScoreBetModel, UserModel
from yak_server.helpers.authentification import encode_bearer_token
from yak_server.helpers.group_position import create_group_position
from yak_server.helpers.logging import (
    logged_in_successfully,
    modify_password_successfully,
    signed_up_successfully,
)

from .utils.auth_utils import is_admin_authentificated, is_authentificated
from .utils.constants import GLOBAL_ENDPOINT, VERSION
from .utils.errors import (
    InvalidCredentials,
    NameAlreadyExists,
    UserNotFound,
)
from .utils.flask_utils import success_response
from .utils.schemas import SCHEMA_LOGIN, SCHEMA_PATCH_USER, SCHEMA_SIGNUP
from .utils.validation import validate_body

auth = Blueprint("auth", __name__)

logger = logging.getLogger(__name__)


@auth.post(f"/{GLOBAL_ENDPOINT}/{VERSION}/users/login")
@validate_body(schema=SCHEMA_LOGIN)
def login_post():
    body = request.get_json()
    user = UserModel.authenticate(body["name"], body["password"])

    if not user:
        raise InvalidCredentials

    token = encode_bearer_token(
        sub=user.id,
        expiration_time=timedelta(seconds=current_app.config["JWT_EXPIRATION_TIME"]),
        secret_key=current_app.config["SECRET_KEY"],
    )

    logger.info(logged_in_successfully(user.name))

    return success_response(HTTPStatus.CREATED, {**user.to_user_dict(), "token": token})


@auth.post(f"/{GLOBAL_ENDPOINT}/{VERSION}/users/signup")
@validate_body(schema=SCHEMA_SIGNUP)
def signup_post():
    data = request.get_json()

    # Check existing user in db
    existing_user = UserModel.query.filter_by(name=data["name"]).first()
    if existing_user:
        raise NameAlreadyExists(data["name"])

    # Initialize user and integrate in db
    user = UserModel(data["name"], data["first_name"], data["last_name"], data["password"])
    db.session.add(user)
    db.session.flush()

    # Initialize bets and integrate in db
    db.session.add_all(
        match.bet_type_from_match.value(user_id=user.id, match_id=match.id)
        for match in MatchModel.query.filter(MatchModel.bet_type_from_match != sql.null())
    )
    db.session.flush()

    # Create group position records
    db.session.add_all(create_group_position(ScoreBetModel.query.filter_by(user_id=user.id)))
    db.session.commit()

    token = encode_bearer_token(
        sub=user.id,
        expiration_time=timedelta(seconds=current_app.config["JWT_EXPIRATION_TIME"]),
        secret_key=current_app.config["SECRET_KEY"],
    )

    logger.info(signed_up_successfully(user.name))

    return success_response(HTTPStatus.CREATED, {**user.to_user_dict(), "token": token})


@auth.patch(f"/{GLOBAL_ENDPOINT}/{VERSION}/users/<string:user_id>")
@validate_body(schema=SCHEMA_PATCH_USER)
@is_admin_authentificated
def patch_user(_, user_id):
    body = request.get_json()

    user = UserModel.query.filter_by(id=user_id).first()
    if not user:
        raise UserNotFound

    user.change_password(body["password"])

    db.session.commit()

    logger.info(modify_password_successfully(user.name))

    return success_response(HTTPStatus.OK, user.to_user_dict())


@auth.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/current_user")
@is_authentificated
def current_user(current_user):
    return success_response(HTTPStatus.OK, current_user.to_user_dict())



import logging
from http import HTTPStatus

from flask import Blueprint, request
from sqlalchemy import update
from sqlalchemy.exc import IntegrityError

from yak_server import db
from yak_server.database.models import (
    GroupPositionModel,
    MatchModel,
    ScoreBetModel,
    is_locked,
)
from yak_server.helpers.logging import modify_score_bet_successfully

from .utils.auth_utils import is_authentificated
from .utils.constants import GLOBAL_ENDPOINT, VERSION
from .utils.errors import (
    BetNotFound,
    GroupNotFound,
    LockedScoreBet,
    TeamNotFound,
)
from .utils.flask_utils import success_response
from .utils.schemas import SCHEMA_PATCH_SCORE_BET, SCHEMA_POST_SCORE_BET
from .utils.validation import validate_body

score_bets = Blueprint("score_bets", __name__)

logger = logging.getLogger(__name__)


@score_bets.post(f"/{GLOBAL_ENDPOINT}/{VERSION}/score_bets")
@validate_body(schema=SCHEMA_POST_SCORE_BET)
@is_authentificated
def create_score_bet(user):
    if is_locked(user.name):
        raise LockedScoreBet

    body = request.get_json()

    match = MatchModel(
        team1_id=body["team1"]["id"],
        team2_id=body["team2"]["id"],
        index=body["index"],
        group_id=body["group"]["id"],
    )

    db.session.add(match)
    try:
        db.session.flush()
    except IntegrityError as integrity_error:
        if "FOREIGN KEY (`team1_id`)" in str(integrity_error):
            raise TeamNotFound(team_id=body["team1"]["id"]) from integrity_error
        elif "FOREIGN KEY (`team2_id`)" in str(integrity_error):
            raise TeamNotFound(team_id=body["team2"]["id"]) from integrity_error
        else:
            raise GroupNotFound(group_id=body["group"]["id"]) from integrity_error

    score_bet = ScoreBetModel(
        match_id=match.id,
        user_id=user.id,
        score1=body["team1"].get("score"),
        score2=body["team2"].get("score"),
    )

    db.session.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == body["team1"]["id"],
            GroupPositionModel.user_id == user.id,
        ),
    )
    db.session.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == body["team2"]["id"],
            GroupPositionModel.user_id == user.id,
        ),
    )

    db.session.add(score_bet)
    db.session.commit()

    return success_response(
        HTTPStatus.CREATED,
        {
            "phase": score_bet.match.group.phase.to_dict(),
            "group": score_bet.match.group.to_dict_without_phase(),
            "score_bet": score_bet.to_dict_without_group(),
        },
    )


@score_bets.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/score_bets/<string:bet_id>")
@is_authentificated
def retrieve_score_bet(user, bet_id):
    score_bet = ScoreBetModel.query.filter_by(user_id=user.id, id=bet_id).first()

    if not score_bet:
        raise BetNotFound(bet_id)

    return success_response(
        HTTPStatus.OK,
        {
            "phase": score_bet.match.group.phase.to_dict(),
            "group": score_bet.match.group.to_dict_without_phase(),
            "score_bet": score_bet.to_dict_without_group(),
        },
    )


@score_bets.patch(f"/{GLOBAL_ENDPOINT}/{VERSION}/score_bets/<string:bet_id>")
@validate_body(schema=SCHEMA_PATCH_SCORE_BET)
@is_authentificated
def modify_score_bet(user, bet_id):
    if is_locked(user.name):
        raise LockedScoreBet

    score_bet = ScoreBetModel.query.filter_by(user_id=user.id, id=bet_id).with_for_update().first()

    if not score_bet:
        raise BetNotFound(bet_id)

    def send_response(score_bet):
        return success_response(
            HTTPStatus.OK,
            {
                "phase": score_bet.match.group.phase.to_dict(),
                "group": score_bet.match.group.to_dict_without_phase(),
                "score_bet": score_bet.to_dict_without_group(),
            },
        )

    body = request.get_json()

    if score_bet.score1 == body["team1"]["score"] and score_bet.score2 == body["team2"]["score"]:
        return send_response(score_bet)

    logger.info(
        modify_score_bet_successfully(
            user.name,
            score_bet,
            body["team1"]["score"],
            body["team2"]["score"],
        ),
    )

    db.session.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == score_bet.match.team1_id,
            GroupPositionModel.user_id == user.id,
        ),
    )
    db.session.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == score_bet.match.team2_id,
            GroupPositionModel.user_id == user.id,
        ),
    )

    score_bet.score1 = body["team1"]["score"]
    score_bet.score2 = body["team2"]["score"]
    db.session.commit()

    return send_response(score_bet)


@score_bets.delete(f"/{GLOBAL_ENDPOINT}/{VERSION}/score_bets/<string:bet_id>")
@is_authentificated
def delete_score_bet(user, bet_id):
    if is_locked(user.name):
        raise LockedScoreBet

    score_bet = ScoreBetModel.query.filter_by(id=bet_id, user_id=user.id).first()

    if not score_bet:
        raise BetNotFound(bet_id)

    response_body = {
        "phase": score_bet.match.group.phase.to_dict(),
        "group": score_bet.match.group.to_dict_without_phase(),
        "score_bet": score_bet.to_dict_without_group(),
    }

    db.session.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == score_bet.match.team1_id,
            GroupPositionModel.user_id == user.id,
        ),
    )
    db.session.execute(
        update(GroupPositionModel)
        .values(need_recomputation=True)
        .where(
            GroupPositionModel.team_id == score_bet.match.team2_id,
            GroupPositionModel.user_id == user.id,
        ),
    )

    db.session.delete(score_bet)
    db.session.commit()

    return success_response(HTTPStatus.OK, response_body)



import logging
from http import HTTPStatus
from operator import attrgetter
from uuid import uuid4

from flask import Blueprint, request
from sqlalchemy import and_

from yak_server import db
from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    GroupPositionModel,
    MatchModel,
    PhaseModel,
    ScoreBetModel,
    is_locked,
)
from yak_server.database.query import (
    bets_from_group_code,
    bets_from_phase_code,
    binary_bets_from_phase_code,
)
from yak_server.helpers.group_position import compute_group_rank

from .utils.auth_utils import is_authentificated
from .utils.constants import GLOBAL_ENDPOINT, VERSION
from .utils.errors import (
    GroupNotFound,
    LockedBinaryBet,
    PhaseNotFound,
)
from .utils.flask_utils import success_response
from .utils.schemas import (
    SCHEMA_PUT_BINARY_BETS_BY_PHASE,
)
from .utils.validation import validate_body

bets = Blueprint("bets", __name__)

logger = logging.getLogger(__name__)


@bets.put(f"/{GLOBAL_ENDPOINT}/{VERSION}/binary_bets/phases/<string:phase_code>")
@validate_body(schema=SCHEMA_PUT_BINARY_BETS_BY_PHASE)
@is_authentificated
def create_bet(current_user, phase_code):
    if is_locked(current_user.name):
        raise LockedBinaryBet

    phase = PhaseModel.query.filter_by(code=phase_code).first()

    groups = GroupModel.query.filter(
        and_(
            GroupModel.phase_id == phase.id,
            GroupModel.code != "8",
        ),
    )

    existing_binary_bets = (
        current_user.binary_bets.filter(
            MatchModel.group_id.in_(map(attrgetter("id"), groups)),
        )
        .join(BinaryBetModel.match)
        .join(MatchModel.group)
        .order_by(GroupModel.index, MatchModel.index)
    )

    body = request.get_json()

    new_matches = []
    new_binary_bets = []

    for bet in body:
        group_id = bet["group"]["id"]
        team1_id = bet["team1"]["id"]
        team2_id = bet["team2"]["id"]
        index = bet["index"]

        match = MatchModel.query.filter_by(
            group_id=group_id,
            index=index,
            team1_id=team1_id,
            team2_id=team2_id,
        ).first()

        if not match:
            match = MatchModel(
                id=str(uuid4()),
                group_id=group_id,
                index=index,
                team1_id=team1_id,
                team2_id=team2_id,
            )

        new_matches.append(match)

        binary_bet = BinaryBetModel.query.filter_by(
            match_id=match.id,
            user_id=current_user.id,
        ).first()

        if not binary_bet:
            binary_bet = BinaryBetModel(match_id=match.id, user_id=current_user.id)

        binary_bet.is_one_won = bet["is_one_won"]

        new_binary_bets.append(binary_bet)

    db.session.add_all(new_matches)
    db.session.flush()

    db.session.add_all(new_binary_bets)
    db.session.flush()

    for bet in existing_binary_bets:
        if bet.id not in map(attrgetter("id"), new_binary_bets):
            db.session.delete(bet)

    db.session.commit()

    phase, groups, binary_bets = binary_bets_from_phase_code(
        current_user,
        phase_code,
    )

    return success_response(
        HTTPStatus.OK,
        {
            "phase": phase.to_dict(),
            "groups": [group.to_dict_without_phase() for group in groups],
            "binary_bets": [binary_bet.to_dict_with_group_id() for binary_bet in binary_bets],
        },
    )


@bets.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/bets")
@is_authentificated
def get_all_bets(current_user):
    binary_bets_query = (
        current_user.binary_bets.join(BinaryBetModel.match)
        .join(MatchModel.group)
        .order_by(GroupModel.index, MatchModel.index)
    )

    score_bets_query = (
        current_user.score_bets.join(ScoreBetModel.match)
        .join(MatchModel.group)
        .order_by(GroupModel.index, MatchModel.index)
    )

    group_query = GroupModel.query.order_by(GroupModel.index)

    phase_query = PhaseModel.query.order_by(PhaseModel.index)

    return success_response(
        HTTPStatus.OK,
        {
            "phases": [phase.to_dict() for phase in phase_query],
            "groups": [group.to_dict_with_phase_id() for group in group_query],
            "score_bets": [score_bet.to_dict_with_group_id() for score_bet in score_bets_query],
            "binary_bets": [binary_bet.to_dict_with_group_id() for binary_bet in binary_bets_query],
        },
    )


@bets.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/bets/phases/<string:phase_code>")
@is_authentificated
def get_bets_by_phase(current_user, phase_code):
    phase, groups, score_bets, binary_bets = bets_from_phase_code(
        current_user,
        phase_code,
    )

    if not phase:
        raise PhaseNotFound(phase_code)

    return success_response(
        HTTPStatus.OK,
        {
            "phase": phase.to_dict(),
            "groups": [group.to_dict_without_phase() for group in groups],
            "score_bets": [score_bet.to_dict_with_group_id() for score_bet in score_bets],
            "binary_bets": [binary_bet.to_dict_with_group_id() for binary_bet in binary_bets],
        },
    )


@bets.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/bets/groups/<string:group_code>")
@is_authentificated
def group_get(current_user, group_code):
    group, score_bets, binary_bets = bets_from_group_code(current_user, group_code)

    if not group:
        raise GroupNotFound(group_code)

    return success_response(
        HTTPStatus.OK,
        {
            "phase": group.phase.to_dict(),
            "group": group.to_dict_without_phase(),
            "binary_bets": [bet.to_dict_without_group() for bet in binary_bets],
            "score_bets": [bet.to_dict_without_group() for bet in score_bets],
        },
    )


@bets.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/bets/groups/rank/<string:group_code>")
@is_authentificated
def group_result_get(current_user, group_code):
    return success_response(
        HTTPStatus.OK,
        get_group_rank_with_code(current_user, group_code),
    )


def get_group_rank_with_code(user, group_code):
    group = GroupModel.query.filter_by(code=group_code).first()

    if not group:
        raise GroupNotFound(group_code)

    group_rank = GroupPositionModel.query.filter_by(group_id=group.id, user_id=user.id)

    def send_response(group, group_rank):
        return {
            "phase": group.phase.to_dict(),
            "group": group.to_dict_without_phase(),
            "group_rank": sorted(
                [group_position.to_dict() for group_position in group_rank],
                key=lambda team: (
                    team["points"],
                    team["goals_difference"],
                    team["goals_for"],
                ),
                reverse=True,
            ),
        }

    if not any(group_position.need_recomputation for group_position in group_rank):
        return send_response(group, group_rank)

    score_bets = user.score_bets.filter(MatchModel.group_id == group.id).join(ScoreBetModel.match)

    group_rank = compute_group_rank(group_rank, score_bets)

    db.session.commit()

    return send_response(group, group_rank)



from datetime import datetime
from enum import Enum
from uuid import uuid4

from dateutil import parser
from flask import current_app, url_for
from sqlalchemy import CheckConstraint
from sqlalchemy import Enum as SqlEnum
from werkzeug.security import check_password_hash, generate_password_hash

from yak_server import db


class UserModel(db.Model):
    __tablename__ = "user"
    id = db.Column(
        db.String(100),
        primary_key=True,
        nullable=False,
        default=lambda: str(uuid4()),
    )
    name = db.Column(db.String(100), unique=True, nullable=False)
    first_name = db.Column(db.String(100), nullable=False)
    last_name = db.Column(db.String(100), nullable=False)
    password = db.Column(db.String(100), nullable=False)
    number_match_guess = db.Column(
        db.Integer,
        CheckConstraint("number_match_guess>=0"),
        nullable=False,
        default=0,
    )
    number_score_guess = db.Column(
        db.Integer,
        CheckConstraint("number_score_guess>=0"),
        nullable=False,
        default=0,
    )
    number_qualified_teams_guess = db.Column(
        db.Integer,
        CheckConstraint("number_qualified_teams_guess>=0"),
        nullable=False,
        default=0,
    )
    number_first_qualified_guess = db.Column(
        db.Integer,
        CheckConstraint("number_first_qualified_guess>=0"),
        nullable=False,
        default=0,
    )
    number_quarter_final_guess = db.Column(
        db.Integer,
        CheckConstraint("number_quarter_final_guess>=0"),
        nullable=False,
        default=0,
    )
    number_semi_final_guess = db.Column(
        db.Integer,
        CheckConstraint("number_semi_final_guess>=0"),
        nullable=False,
        default=0,
    )
    number_final_guess = db.Column(
        db.Integer,
        CheckConstraint("number_final_guess>=0"),
        nullable=False,
        default=0,
    )
    number_winner_guess = db.Column(
        db.Integer,
        CheckConstraint("number_winner_guess>=0"),
        nullable=False,
        default=0,
    )

    points = db.Column(
        db.Float,
        CheckConstraint("points>=0"),
        nullable=False,
        default=0,
    )

    score_bets = db.relationship(
        "ScoreBetModel",
        back_populates="user",
        lazy="dynamic",
        cascade="all, delete",
        passive_deletes=True,
    )

    binary_bets = db.relationship(
        "BinaryBetModel",
        back_populates="user",
        lazy="dynamic",
        cascade="all, delete",
        passive_deletes=True,
    )

    def __init__(self, name, first_name, last_name, password) -> None:
        self.name = name
        self.first_name = first_name
        self.last_name = last_name
        self.password = generate_password_hash(password, method="sha256")

    @classmethod
    def authenticate(cls, name, password):
        user = cls.query.filter_by(name=name).first()
        if not user or not check_password_hash(user.password, password):
            return None

        return user

    def change_password(self, new_password) -> None:
        self.password = generate_password_hash(new_password, method="sha256")

    def to_user_dict(self):
        return {
            "id": self.id,
            "name": self.name,
        }

    def to_result_dict(self):
        return {
            "first_name": self.first_name,
            "last_name": self.last_name,
            "full_name": f"{self.first_name} {self.last_name}",
            "number_match_guess": self.number_match_guess,
            "number_score_guess": self.number_score_guess,
            "number_qualified_teams_guess": self.number_qualified_teams_guess,
            "number_first_qualified_guess": self.number_first_qualified_guess,
            "number_quarter_final_guess": self.number_quarter_final_guess,
            "number_semi_final_guess": self.number_semi_final_guess,
            "number_final_guess": self.number_final_guess,
            "number_winner_guess": self.number_winner_guess,
            "points": round(self.points, 3),
        }


def is_locked(user_name):
    locked_date = parser.parse(current_app.config["LOCK_DATETIME"])
    return user_name != "admin" and datetime.now() > locked_date


class ScoreBetModel(db.Model):
    __tablename__ = "score_bet"
    id = db.Column(
        db.String(100),
        primary_key=True,
        nullable=False,
        default=lambda: str(uuid4()),
    )
    user_id = db.Column(
        db.String(100),
        db.ForeignKey("user.id", ondelete="CASCADE"),
        nullable=False,
    )
    user = db.relationship("UserModel", back_populates="score_bets")

    match_id = db.Column(db.String(100), db.ForeignKey("match.id"), nullable=False)
    match = db.relationship("MatchModel", back_populates="score_bets")

    score1 = db.Column(db.Integer, CheckConstraint("score1>=0"), default=None)
    score2 = db.Column(db.Integer, CheckConstraint("score2>=0"), default=None)

    def is_invalid(self) -> bool:
        return None in (self.score1, self.score2)

    def is_valid(self) -> bool:
        return not self.is_invalid()

    def is_1_win(self) -> bool:
        return self.is_valid() and self.score1 > self.score2

    def is_draw(self) -> bool:
        return self.is_valid() and self.score1 == self.score2

    def is_2_win(self) -> bool:
        return self.is_valid() and self.score1 < self.score2

    def is_same_results(self, other) -> bool:
        return (
            (self.is_1_win() and other.is_1_win())
            or (self.is_draw() and other.is_draw())
            or (self.is_2_win() and other.is_2_win())
        )

    def is_same_scores(self, other) -> bool:
        return (
            self.is_valid()
            and other.is_valid()
            and self.score1 == other.score1
            and self.score2 == other.score2
        )

    def to_dict_with_group_id(self):
        return {
            "id": self.id,
            "index": self.match.index,
            "locked": is_locked(self.user.name),
            "group": {"id": self.match.group_id},
            "team1": {**self.match.team1.to_dict(), "score": self.score1},
            "team2": {**self.match.team2.to_dict(), "score": self.score2},
        }

    def to_dict_without_group(self):
        return {
            "id": self.id,
            "index": self.match.index,
            "locked": is_locked(self.user.name),
            "team1": {**self.match.team1.to_dict(), "score": self.score1},
            "team2": {**self.match.team2.to_dict(), "score": self.score2},
        }


class BinaryBetModel(db.Model):
    __tablename__ = "binary_bet"
    id = db.Column(
        db.String(100),
        primary_key=True,
        nullable=False,
        default=lambda: str(uuid4()),
    )
    user_id = db.Column(
        db.String(100),
        db.ForeignKey("user.id", ondelete="CASCADE"),
        nullable=False,
    )
    user = db.relationship("UserModel", back_populates="binary_bets")

    match_id = db.Column(db.String(100), db.ForeignKey("match.id"), nullable=False)
    match = db.relationship("MatchModel", back_populates="binary_bets")

    is_one_won = db.Column(db.Boolean, default=None)

    def bet_from_is_one_won(self):
        if self.is_one_won is None:
            return (None, None)

        return (self.is_one_won, not self.is_one_won)

    def to_dict_with_group_id(self):
        bet_results = self.bet_from_is_one_won()

        return {
            "id": self.id,
            "index": self.match.index,
            "locked": is_locked(self.user.name),
            "group": {"id": self.match.group_id},
            "team1": {**self.match.team1.to_dict(), "won": bet_results[0]},
            "team2": {**self.match.team2.to_dict(), "won": bet_results[1]},
        }

    def to_dict_without_group(self):
        bet_results = self.bet_from_is_one_won()

        return {
            "id": self.id,
            "index": self.match.index,
            "locked": is_locked(self.user.name),
            "team1": {**self.match.team1.to_dict(), "won": bet_results[0]},
            "team2": {**self.match.team2.to_dict(), "won": bet_results[1]},
        }


class BetMapping(Enum):
    SCORE_BET = ScoreBetModel
    BINARY_BET = BinaryBetModel


class MatchModel(db.Model):
    __tablename__ = "match"
    id = db.Column(
        db.String(100),
        primary_key=True,
        nullable=False,
        default=lambda: str(uuid4()),
    )

    group_id = db.Column(db.String(100), db.ForeignKey("group.id"), nullable=False)
    group = db.relationship("GroupModel", foreign_keys=group_id, backref="matches")

    index = db.Column(db.Integer, nullable=False)

    team1_id = db.Column(db.String(100), db.ForeignKey("team.id"), nullable=False)
    team1 = db.relationship("TeamModel", foreign_keys=team1_id, backref="match1")

    team2_id = db.Column(db.String(100), db.ForeignKey("team.id"), nullable=False)
    team2 = db.relationship("TeamModel", foreign_keys=team2_id, backref="match2")

    score_bets = db.relationship("ScoreBetModel", back_populates="match")
    binary_bets = db.relationship("BinaryBetModel", back_populates="match")

    bet_type_from_match = db.Column(SqlEnum(BetMapping), default=None)


class TeamModel(db.Model):
    __tablename__ = "team"
    id = db.Column(
        db.String(100),
        primary_key=True,
        nullable=False,
        default=lambda: str(uuid4()),
    )
    code = db.Column(db.String(10), unique=True, nullable=False)
    description = db.Column(db.String(100), unique=True, nullable=False)
    flag_url = db.Column(db.String(100))

    def to_dict(self):
        return {
            "id": self.id,
            "code": self.code,
            "description": self.description,
            "flag": {"url": url_for("team.retrieve_team_flag", team_id=self.id)},
        }


class GroupModel(db.Model):
    __tablename__ = "group"
    id = db.Column(
        db.String(100),
        primary_key=True,
        nullable=False,
        default=lambda: str(uuid4()),
    )
    code = db.Column(db.String(1), primary_key=True, unique=True, nullable=False)
    description = db.Column(db.String(100), unique=True, nullable=False)
    index = db.Column(db.Integer, nullable=False)

    phase_id = db.Column(db.String(100), db.ForeignKey("phase.id"), nullable=False)
    phase = db.relationship("PhaseModel", backref="groups")

    def to_dict_with_phase_id(self):
        return {
            "id": self.id,
            "code": self.code,
            "phase": {"id": self.phase_id},
            "description": self.description,
        }

    def to_dict_without_phase(self):
        return {
            "id": self.id,
            "code": self.code,
            "description": self.description,
        }


class PhaseModel(db.Model):
    __tablename__ = "phase"
    id = db.Column(
        db.String(100),
        primary_key=True,
        nullable=False,
        default=lambda: str(uuid4()),
    )
    code = db.Column(db.String(10), primary_key=True, unique=True, nullable=False)
    description = db.Column(db.String(100), nullable=False)
    index = db.Column(db.Integer, nullable=False)

    def to_dict(self):
        return {
            "id": self.id,
            "code": self.code,
            "description": self.description,
        }


class GroupPositionModel(db.Model):
    __tablename__ = "group_position"
    id = db.Column(
        db.String(100),
        primary_key=True,
        nullable=False,
        default=lambda: str(uuid4()),
    )
    won = db.Column(db.Integer, CheckConstraint("won>=0"), nullable=False, default=0)
    drawn = db.Column(db.Integer, CheckConstraint("drawn>=0"), nullable=False, default=0)
    lost = db.Column(db.Integer, CheckConstraint("lost>=0"), nullable=False, default=0)
    goals_for = db.Column(db.Integer, CheckConstraint("goals_for>=0"), nullable=False, default=0)
    goals_against = db.Column(
        db.Integer,
        CheckConstraint("goals_against>=0"),
        nullable=False,
        default=0,
    )
    need_recomputation = db.Column(db.Boolean, nullable=False, default=False)

    user_id = db.Column(
        db.String(100),
        db.ForeignKey("user.id", ondelete="CASCADE"),
        nullable=False,
    )

    team_id = db.Column(
        db.String(100),
        db.ForeignKey("team.id"),
        nullable=False,
    )
    team = db.relationship("TeamModel")

    group_id = db.Column(
        db.String(100),
        db.ForeignKey("group.id"),
        nullable=False,
    )

    def to_dict(self):
        return {
            "team": self.team.to_dict(),
            "played": self.won + self.drawn + self.lost,
            "won": self.won,
            "drawn": self.drawn,
            "lost": self.lost,
            "goals_for": self.goals_for,
            "goals_against": self.goals_against,
            "goals_difference": self.goals_for - self.goals_against,
            "points": self.won * 3 + self.drawn,
        }



from functools import wraps

from flask import current_app
from jwt import ExpiredSignatureError

from yak_server.database.models import UserModel
from yak_server.helpers.authentification import decode_bearer_token

from .result import ExpiredToken, InvalidToken, UnauthorizedAccessToAdminAPI
from .schema import User

NUMBER_ELEMENTS_IN_AUTHORIZATION = 2


def is_authentificated(f):
    @wraps(f)
    def _verify(*args, **kwargs):
        auth_headers = kwargs["info"].context["request"].headers.get("Authorization", "").split()

        if len(auth_headers) != NUMBER_ELEMENTS_IN_AUTHORIZATION or auth_headers[0] != "Bearer":
            return InvalidToken()

        token = auth_headers[1]
        try:
            data = decode_bearer_token(token, current_app.config["SECRET_KEY"])
        except ExpiredSignatureError:
            return ExpiredToken()
        except Exception:
            return InvalidToken()

        user = UserModel.query.filter_by(id=data["sub"]).first()
        if not user:
            return InvalidToken()

        kwargs["info"].user = User.from_instance(instance=user)

        return f(*args, **kwargs)

    return _verify


def is_admin_authentificated(f):
    @wraps(f)
    def _verify(*args, **kwargs):
        if kwargs["info"].user.pseudo != "admin":
            return UnauthorizedAccessToAdminAPI()

        return f(*args, **kwargs)

    return _verify



from http import HTTPStatus

from flask import Blueprint

from yak_server.database.models import PhaseModel

from .utils.auth_utils import is_authentificated
from .utils.constants import GLOBAL_ENDPOINT, VERSION
from .utils.errors import PhaseNotFound
from .utils.flask_utils import success_response

phase = Blueprint("phase", __name__)


@phase.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/phases")
@is_authentificated
def retrieve_all_phases(user):
    return success_response(
        HTTPStatus.OK,
        [phase.to_dict() for phase in PhaseModel.query.order_by(PhaseModel.index)],
    )


@phase.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/phases/<string:phase_id>")
@is_authentificated
def retrieve_by_phase_id(user, phase_id):
    phase = PhaseModel.query.filter_by(id=phase_id).first()

    if not phase:
        raise PhaseNotFound(phase_id)

    return success_response(HTTPStatus.OK, phase.to_dict())



import json
import os
import sys
from configparser import ConfigParser

if sys.version_info >= (3, 9):
    from importlib import resources
else:
    import importlib_resources as resources

from pathlib import Path

from .helpers.rules import RULE_MAPPING


class RuleNotDefined(Exception):
    def __init__(self, rule_id) -> None:
        super().__init__(f"Rule not defined: {rule_id}")


def compute_database_uri(mysql_user_name, mysql_password, mysql_port, mysql_db):
    return f"mysql+pymysql://{mysql_user_name}:{mysql_password}@localhost:{mysql_port}/{mysql_db}"


def get_mysql_config():
    mysql_config = {
        "MYSQL_USER_NAME": os.environ["MYSQL_USER_NAME"],
        "MYSQL_PASSWORD": os.environ["MYSQL_PASSWORD"],
        "MYSQL_PORT": os.environ.get("MYSQL_PORT", 3306),
        "MYSQL_DB": os.environ["MYSQL_DB"],
        "SQLALCHEMY_TRACK_MODIFICATIONS": False,
    }

    mysql_config["SQLALCHEMY_DATABASE_URI"] = compute_database_uri(
        mysql_config["MYSQL_USER_NAME"],
        mysql_config["MYSQL_PASSWORD"],
        mysql_config["MYSQL_PORT"],
        mysql_config["MYSQL_DB"],
    )

    return mysql_config


def get_jwt_config():
    return {
        "SECRET_KEY": os.environ["JWT_SECRET_KEY"],
        "JWT_EXPIRATION_TIME": int(os.environ["JWT_EXPIRATION_TIME"]),
    }


def get_yak_config():
    with resources.as_file(
        resources.files("yak_server") / "data" / os.environ["COMPETITION"],
    ) as path:
        DATA_FOLDER = path

    config = ConfigParser()
    config.read(f"{DATA_FOLDER}/config.ini")

    with resources.as_file(
        resources.files("yak_server") / "data" / os.environ["COMPETITION"],
    ) as path:
        DATA_FOLDER = path

    config = ConfigParser()
    config.read(f"{DATA_FOLDER}/config.ini")

    RULES = {}

    for rule_file in Path(f"{DATA_FOLDER}/rules").glob("*.json"):
        rule_id = rule_file.stem

        if rule_id not in RULE_MAPPING:
            raise RuleNotDefined(rule_id)

        with rule_file.open() as rule_content:
            RULES[rule_id] = json.loads(rule_content.read())

    return {
        # SQL Alchemy features
        "LOCK_DATETIME": config.get("locking", "datetime"),
        "BASE_CORRECT_RESULT": config.getint("points", "base_correct_result"),
        "MULTIPLYING_FACTOR_CORRECT_RESULT": config.getint(
            "points",
            "multiplying_factor_correct_result",
        ),
        "BASE_CORRECT_SCORE": config.getint("points", "base_correct_score"),
        "MULTIPLYING_FACTOR_CORRECT_SCORE": config.getint(
            "points",
            "multiplying_factor_correct_score",
        ),
        "TEAM_QUALIFIED": config.getint("points", "team_qualified"),
        "FIRST_TEAM_QUALIFIED": config.getint("points", "first_team_qualified"),
        "DATA_FOLDER": DATA_FOLDER,
        "RULES": RULES,
    }


def get_config():
    return {**get_mysql_config(), **get_jwt_config(), **get_yak_config()}



from http import HTTPStatus

from flask import Blueprint, current_app

from yak_server.helpers.rules import RULE_MAPPING

from .utils.auth_utils import is_authentificated
from .utils.constants import GLOBAL_ENDPOINT, VERSION
from .utils.errors import RuleNotFound
from .utils.flask_utils import success_response

rules = Blueprint("rules", __name__)


@rules.post(f"/{GLOBAL_ENDPOINT}/{VERSION}/rules/<string:rule_id>")
@is_authentificated
def execute_rule(user, rule_id):
    if rule_id not in current_app.config["RULES"]:
        raise RuleNotFound(rule_id)

    RULE_MAPPING[rule_id](user, current_app.config["RULES"][rule_id])

    return success_response(HTTPStatus.OK, None)



import click

from yak_server import create_app

from .database import (
    backup_database,
    create_admin,
    create_database,
    delete_database,
    drop_database,
    initialize_database,
)


@click.command()
def create():
    """Create all database tables"""
    app = create_app()
    with app.app_context():
        create_database(app)


@click.command()
def init():
    """Initialize database"""
    app = create_app()
    with app.app_context():
        initialize_database(app)


@click.command()
def drop():
    """Drop all tables"""
    app = create_app()
    with app.app_context():
        drop_database(app)


@click.command()
def delete():
    """Delete all records"""
    app = create_app()
    with app.app_context():
        delete_database(app)


@click.command()
def admin():
    """Create admin account in database"""
    app = create_app()
    with app.app_context():
        create_admin(app)


@click.command()
def backup():
    """Backup database in a sql file"""
    app = create_app()
    with app.app_context():
        backup_database(app)


@click.group()
def db():
    pass


db.add_command(create)
db.add_command(init)
db.add_command(drop)
db.add_command(delete)
db.add_command(admin)
db.add_command(backup)


@click.group()
def main():
    pass


main.add_command(db)



SCHEMA_POST_SCORE_BET = {
    "type": "object",
    "properties": {
        "index": {"type": "integer", "minimum": 1},
        "team1": {
            "type": "object",
            "properties": {
                "id": {"format": "uuid"},
                "score": {
                    "oneOf": [
                        {
                            "type": "integer",
                            "minimum": 0,
                        },
                        {"type": "null"},
                    ],
                },
            },
            "required": ["id"],
        },
        "team2": {
            "type": "object",
            "properties": {
                "id": {"format": "uuid"},
                "score": {
                    "oneOf": [
                        {
                            "type": "integer",
                            "minimum": 0,
                        },
                        {"type": "null"},
                    ],
                },
            },
            "required": ["id"],
        },
        "group": {
            "type": "object",
            "properties": {
                "id": {"format": "uuid"},
            },
            "required": ["id"],
        },
    },
    "required": ["team1", "team2", "group", "index"],
}

SCHEMA_PATCH_SCORE_BET = {
    "type": "object",
    "properties": {
        "team1": {
            "type": "object",
            "properties": {
                "score": {
                    "oneOf": [
                        {
                            "type": "integer",
                            "minimum": 0,
                        },
                        {"type": "null"},
                    ],
                },
            },
            "required": ["score"],
        },
        "team2": {
            "type": "object",
            "properties": {
                "score": {
                    "oneOf": [
                        {
                            "type": "integer",
                            "minimum": 0,
                        },
                        {"type": "null"},
                    ],
                },
            },
            "required": ["score"],
        },
    },
    "required": ["team1", "team2"],
}

SCHEMA_POST_BINARY_BET = {
    "type": "object",
    "properties": {
        "is_one_won": {
            "oneOf": [{"type": "boolean"}, {"type": "null"}],
        },
        "index": {"type": "integer", "minimum": 1},
        "team1": {
            "type": "object",
            "properties": {
                "id": {"format": "uuid"},
            },
            "required": ["id"],
        },
        "team2": {
            "type": "object",
            "properties": {
                "id": {"format": "uuid"},
            },
            "required": ["id"],
        },
        "group": {
            "type": "object",
            "properties": {
                "id": {"format": "uuid"},
            },
            "required": ["id"],
        },
    },
    "required": ["index", "team1", "team2", "group"],
}

SCHEMA_PATCH_BINARY_BET = {
    "type": "object",
    "properties": {
        "is_one_won": {
            "oneOf": [{"type": "boolean"}, {"type": "null"}],
        },
    },
    "required": ["is_one_won"],
}

SCHEMA_SIGNUP = {
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "first_name": {"type": "string"},
        "last_name": {"type": "string"},
        "password": {"type": "string"},
    },
    "required": ["name", "first_name", "last_name", "password"],
}


SCHEMA_LOGIN = {
    "type": "object",
    "properties": {"name": {"type": "string"}, "password": {"type": "string"}},
    "required": ["name", "password"],
}


SCHEMA_PATCH_USER = {
    "type": "object",
    "properties": {"password": {"type": "string"}},
    "required": ["password"],
}


SCHEMA_PUT_BINARY_BETS_BY_PHASE = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "is_one_won": {
                "oneOf": [
                    {
                        "type": "boolean",
                    },
                    {
                        "type": "null",
                    },
                ],
            },
            "index": {"type": "integer"},
            "group": {
                "type": "object",
                "properties": {"id": {"format": "uuid"}},
                "required": ["id"],
            },
            "team1": {
                "type": "object",
                "properties": {"id": {"format": "uuid"}},
                "required": ["id"],
            },
            "team2": {
                "type": "object",
                "properties": {"id": {"format": "uuid"}},
                "required": ["id"],
            },
        },
        "required": ["is_one_won", "index", "group", "team1", "team2"],
    },
}



INVALID_CREDENTIALS_MESSAGE = "Invalid credentials"
UNAUTHORIZED_ACCESS_TO_ADMIN_API_MESSAGE = "Unauthorized access to admin API"
INVALID_TOKEN_MESSAGE = "Invalid token, authentication required"
EXPIRED_TOKEN_MESSAGE = "Expired token, reauthentication required"
LOCKED_SCORE_BET_MESSAGE = "Cannot modify score bet, lock date is exceeded"
LOCKED_BINARY_BET_MESSAGE = "Cannot modify binary bet, lock date is exceeded"


def name_already_exists_message(user_name: str) -> str:
    return f"Name already exists: {user_name}"



import strawberry
from flask import current_app
from strawberry.extensions import MaskErrors, QueryDepthLimiter

from .mutation import Mutation
from .query import Query


def mask_errors(_) -> bool:
    return not current_app.config["DEBUG"]


schema = strawberry.Schema(
    query=Query,
    mutation=Mutation,
    extensions=[MaskErrors(should_mask_error=mask_errors), QueryDepthLimiter(max_depth=6)],
)



from http import HTTPStatus

from flask import Blueprint

from yak_server.database.models import GroupModel, PhaseModel

from .utils.auth_utils import is_authentificated
from .utils.constants import GLOBAL_ENDPOINT, VERSION
from .utils.errors import GroupNotFound, PhaseNotFound
from .utils.flask_utils import success_response

groups = Blueprint("group", __name__)


@groups.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/groups")
@is_authentificated
def get_groups(current_user):
    group_query = GroupModel.query.order_by(GroupModel.index)
    groups = [group.to_dict_with_phase_id() for group in group_query]

    phase_query = PhaseModel.query.order_by(PhaseModel.index)
    phases = [phase.to_dict() for phase in phase_query]

    return success_response(
        HTTPStatus.OK,
        {
            "phases": phases,
            "groups": groups,
        },
    )


@groups.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/groups/<string:group_code>")
@is_authentificated
def get_group_by_code(current_user, group_code):
    group = GroupModel.query.filter_by(code=group_code).first()

    if not group:
        raise GroupNotFound(group_code)

    return success_response(
        HTTPStatus.OK,
        {
            "phase": group.phase.to_dict(),
            "group": group.to_dict_without_phase(),
        },
    )


@groups.get(f"/{GLOBAL_ENDPOINT}/{VERSION}/groups/phases/<string:phase_code>")
@is_authentificated
def get_groups_by_phase_code(current_user, phase_code):
    phase = PhaseModel.query.filter_by(code=phase_code).first()

    if not phase:
        raise PhaseNotFound(phase_code)

    group_query = GroupModel.query.order_by(GroupModel.index).filter_by(phase_id=phase.id)
    groups = [group.to_dict_without_phase() for group in group_query]

    return success_response(
        HTTPStatus.OK,
        {
            "phase": phase.to_dict(),
            "groups": groups,
        },
    )
